// Not compiled, but compiled at runtime by Tavern

syntax = "proto3";

package helloworld.v2;

message HelloRequest {
  string name = 1;
}

message HelloReply {
  string message = 1;
}

service Greeter {
  rpc SayHello (HelloRequest) returns (HelloReply) {}
}
