// Pre compiled and checked into the repo so it can be imported by Tavern at runtime

syntax = "proto3";

package helloworld.v1;

message HelloRequest {
  string name = 1;
}

message HelloReply {
  string message = 1;
}

service Greeter {
  rpc SayHello (HelloRequest) returns (HelloReply) {}
}
